// midi.cpp

/* The code in this file is based on the following document:
 * Standard MIDI-File Format Spec. 1.1
 * which is distributed by The International MIDI Association
 * However, some stuff that didn't seem important was not included in this
 * MIDI interpreter.
 */

#include <stdio.h>
#include <assert.h>
#include <iostream.h>
#include <stdlib.h>
#include <alloc.h>

#include "midi.hpp"

midiChunk::midiChunk() {
	data=0;
	len=0;
	id[0]=id[4]=0;
}

midiChunk::~midiChunk() {
	if(data!=0)
		farfree(data);
}

void midiChunk::read(FILE* in) {
	long l;
	unsigned r;
	pbyte pd;
	fread(id,1,4,in);
	fread(&len,4,1,in);
	swapbytes(len);
	data = (pbyte)farmalloc(len);
	if(data==0) {
		cout<<"Not enough memory!\n";
		abort();
	}
	l=len;
	pd=data;
	while(l>0) {
		r = fread(pd,1,(l>64000 ? 64000 : l),in);
		pd += r;
		l -= r;
	}
}

word midiHeader::format() {
	// returns midi format (0, 1, or 2), or -1 on error
	if(len==0)
		return -1;
	if(data[0]!=0)
		return -1;
	if(data[1]>2)
		return -1;
	return (int)data[1];
}

word midiHeader::ntracks() {
	// returns number of tracks, -1 on error
	word d;
	if(len==0)
		return -1;
	d = *(pword)(data+2);
	swapbytes(d);
	return d;
}

word midiHeader::division() {
	word div;
	if(len==0) return -1;
	div = *(pword)(data+4);
	swapbytes(div);
	return div;
}

bool midiHeader::verify_type() {
	if(id[0]=='M' && id[1]=='T' && id[2]=='h' && id[3]=='d')
		return true;
	return false;
}

inline void swapbytes(dword& dw) {
	dw = (dw<<24) | (dw>>24) | ((dw&0x00ff0000)>>8) | ((dw&0x0000ff00)<<8);
}

inline void swapbytes(word& w) {
	w = (w<<8) | (w>>8);
}

dword readvarlen(pbyte& b, long& bytes_read) {
// reads a variable-length data item and advances the pointer accordingly
	dword val;
	byte c;
	bytes_read++;
	if((val = *(b++)) & 0x80) {
		val &= 0x7f;
		do {
			val = (val<<7) | ((c = *(b++)) & 0x7f);
			bytes_read++;
		} while(c&0x80);
	}
	return val;
}

midiEvent::midiEvent(pbyte& d, long& bytes_read, byte run_stat) {
// parses a midi event and advances the pointer
	dword dw;
	running = run_stat;
	delta = readvarlen(d,bytes_read);
	cmd = *(d++);
	bytes_read++;

	// first, skip sysex msgs and meta-events
	if(cmd==0xf0 || cmd==0xf7) {	// sysex msg - ignore it
		ignore_me = true;
		dw=readvarlen(d,bytes_read);
		d += dw; // skip entire sysex
		bytes_read+=dw;
		return;
	}
	if(cmd==0xff) {	 // meta-event
		cmd = *(d++); // what type of meta-event?
		bytes_read++;
		switch(cmd) {
			case 0x51:	// tempo change
				// high byte in note, mid byte in chan, low byte in vol
				ignore_me = false;
				assert((*(d++)) == 3);// otherwise the command is coded badly
				note = *(d++);
				chan = *(d++);
				volume = *(d++);
				bytes_read++;
				break;
			default:	// anything else is ignored
				ignore_me = true;
				dw = readvarlen(d,bytes_read);
				d+=dw; // skip entire meta-event
				bytes_read+=dw;
				break;
		}
		return;
	}

	// now, parse the ordinary events
	ignore_me = false;
	if((cmd&0x80) == 0) { // check for running status
		d--; // push back the command - it's actually a note number.
		bytes_read--;
		cmd = running; // load command from running status buffer.
	}
	switch(cmd&0xf0) {
		case 0xd0:	// ignore this command and skip one byte
			ignore_me = true;
			running = cmd;
			d++;
			bytes_read++;
			break;
		case 0xc0:	// prog chng command - note contains new prog
			running = cmd;	// save running status
			chan = cmd&0x0f;
			cmd &= 0xf0;
			note = *(d++);
			bytes_read++;
			return;
		case 0x90:	// note on command
			running = cmd;	// save running status
			chan = cmd&0x0f;
			cmd &= 0xf0;
			note = *(d++);
			volume = *(d++);
			if(volume==0) // some midi files use vol 0 to indicate note off
				cmd = 0x80;
			bytes_read+=2;
			return;
		case 0x80:	// note off command (volume should be ignored?)
			running = cmd;	// save running status
			chan = cmd&0x0f;
			cmd &= 0xf0;
			note = *(d++);
			volume = *(d++);
			bytes_read+=2;
			return;
		default:	// unknown command - ignored
			running = cmd;
			d+=2;
			bytes_read+=2;
			ignore_me = true;
			return;
	}
}

midiEvent* midiTrack::curEvent()
{
	if(cur_event.delta==-1)
		return 0;
	return &cur_event;
}

midiEvent* midiTrack::nextEvent()
{
	if(cur_pos==0) { // initialization necessary
		if(data==0)  // track not loaded yet
			return 0;
		cur_pos = data;
        bytes_read = 0;
	}
	do {
		if(bytes_read >= len) // no more events
			return 0;
		cur_event = midiEvent(cur_pos,bytes_read,running);
		cur_time += cur_event.delta;
		running = cur_event.running;
	} while(cur_event.ignore_me);
	return &cur_event;
}

bool midiTrack::verify_type() {
	if(id[0]=='M' && id[1]=='T' && id[2]=='r' && id[3]=='k')
		return true;
	return false;
}

bool midiTrack::end() {
	if(bytes_read>=len)
		return true;
	else
		return false;
}

void midiTrack::reset() {
	cur_pos=0;
	bytes_read=0;
	cur_time=0;
	running=0;
	nextEvent();
}

midiTrack* midiFile::track(int trnum) {
	if(_track==0) return 0;
	if(trnum<0 || trnum>=_header.ntracks()) return 0;
	return _track+trnum;
}

void midiFile::read(FILE* file) {
	int i;
	_header.read(file);
	_track = new midiTrack[_header.ntracks()];
	if(_track==0) {
		cout<<"Not enough memory!\n";
		abort();
	}
	for(i=0; i<_header.ntracks(); i++) {
		_track[i].read(file);
		_track[i].nextEvent();
	}
}

bool midiFile::finished() {
	int i;
	if(_header.format()==(word)(-1)) return false;
	for(i=0; i<_header.ntracks(); i++)
		if(_track[i].end()==0)
			return false;
	return true;
}

int midiFile::nexttrack() {
	int i,retval;
	dword min=0x7fffffff;
	if(_header.format()==(word)(-1)) return -1;
	for(i=0; i<_header.ntracks(); i++) {
		if(_track[i].time()<min && _track[i].end()==0) {
			retval=i;
			min=_track[i].time();
		}
	}
	return retval;
}

void midiFile::reset() {
	int i;
	if(_header.format()==(word)(-1)) return;
	for(i=0; i<_header.ntracks(); i++)
		_track[i].reset();
}
